/*
  Solar MPP-Tracker mit Step-Down Wandler
  =======================================

  Realisiert mit Tiny44; Interner RC  mit 8MHz (Teiler/8 inaktiv),
                         Brown-out Level: 4.3V
                         Fuses: Ext=0xFF High=0xDC Low=0xE2 

  D.Schrder 2012

  Version 1.4 vom 12.10.2012
 
 
Pinbelegung:

                    +5V VCC 1  14 GND
                   FET  PB0 2  13 PA0 ADC0 Modulstrom           0..55mV
                    Tx  PB1 3  12 PA1 ADC1 AIN0 Modulspannung   0..1.1V
                  Reset PB3 4  11 PA2 ADC2 AIN1 Sollspannung In (Komparator)
                        PB2 5  10 PA3 ADC3 Modul GND (neg. diff. ADC input).
                        PA7 6  9  PA4 ADC4 
  Sollspannung Out(PWM) PA6 7  8  PA5 ADC5 
             (OC1A)           
             
             
  ADC: Interne 1.1V Referenz. Messung alle 32ms
    Strommessung:      (ADC0 - ADC3)x20  
    Spannungsmessung:  (ADC1 - ADC3)x 1  
    Die Messergebnisse werden mit 9600Baud auf Pin3 (Tx) ausgegeben
    
  PWM: Das Tastverhltnis stellt die Modulsollspannung ein. Dazu wird mit dem eingebauten
    Komparator eine Art Hysteretic Schaltwandler realisiert.
    Umodul und Usoll(aus PWM) auf internen Komparator (Umodul:AIN0->Pin12, Usoll: AIN1->Pin11). 
    Solange Umodul>Usoll:  FET ein, sonst aus.
    Auswertung des Komparators (ACO) in Naked-Interrupt-Routine mit hoher Taktfrequenz (64KHz)   
    Da der Prozessor mit 5V luft, die Referenz aber 1.1V ist, muss nach dem PWM ein RC-Glied 
    und ein Spannungsteiler 1.1 / 5 folgen.

  Step-Down-Wandler:
    An Pin2 wird ber einen 120Ohm Widerstand das Gate eines FET (LogicLevel Typ wie z.B. IRLU024N) angeschlossen. 
    (Schutz des Ports: Portpin kann max. 40mA. Mit 120 Ohm knnen bei 5V nicht mehr flieen.)
    FET: Source auf GND,  Drain ber Spule und Ausgangselko an Modul+. Schnelle Diode mit Anode an Drain, Kathode an Modul+.
    Batterie (Last) kommt parallel zum Ausgangselko. 
    Zustzlich ein Eingangselko an Modul+ und Shunt (nicht Modul-)
    

  Batterie: 
    Aktive Z-Diodenschaltung ber die Batterie. Die begrenzt die Ladeschluspannung nach oben
    und verbrt berflssigen Strom. 
    Vorteil: Der MPP wird weiterhin verfolgt und bei pltzlicher Belastung des Akkus muss 
    er nicht lange gesucht werden.
    Wenn der Akku aus mehreren Zellen besteht kann man ber jeder Zelle so eine Schaltung als Balancer einsetzen.

  Compiler:
    Optimizer auf -Os 
  
*/



#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h> //Fr memset
#include <util/delay.h>

#define U8 unsigned char
#define I8 signed char
#define U16 unsigned int
#define I16 signed int
#define U32 unsigned long
#define I32 signed long

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128


register unsigned char  statusInIrqFunc   asm("r16"); //R16 reservieren fr Statusspeicherung in naked IRQ Routine
register unsigned char  ticCount64us      asm("r17"); //R17 reservieren fr Zhler in naked IRQ Routine

//ALle globale Variablen in einer Struktur
typedef struct{
  U8 ms32;   // 32ms Zhler 0..255 (bis 8s)
  U16 ms100; //100ms Zhler 0..30000  (bis 50 Minuten)
}Tglob;

Tglob g;

//Zwei Median-Filter. Filtern Ausreier raus.
//------------------------------------------------------------------------------------------------------------ 
U16 medianU(U16 value){
//------------------------------------------------------------------------------------------------------------ 
//Speichert eintreffende Werte in einem 3-stufigen FIFO.
//Sortiert die 3 Werte im FIFO nach Gre und liefert den Wert in der Mitte.
//Der Filter liefert immer einen echten Mewerte (also keinen Mittelwert).
//Einzelne Ausreier werden dabei rausgefiltert..
static U16 a,b,c, f1,f2,f3, x;
  f1=f2;
  f2=f3;
  f3=value;
  
  a=f1;
  b=f2;
  c=f3;

  if (a>b){
    x=a;
    a=b;
    b=x; 
  } //here is a<=b

  if (b>c){
    b=c;
    //x=c;
    //c=b;
    //b=x;
  }//here is  b<=c
  if (a>b){
    return a;
    //x=a;
    //a=b;
    //b=x; 
  } //here is a<=b

  return b;
}

//------------------------------------------------------------------------------------------------------------ 
U16 medianI(U16 value){
//------------------------------------------------------------------------------------------------------------ 
static U16 a,b,c, f1,f2,f3, x;
  f1=f2;
  f2=f3;
  f3=value;
  
  a=f1;
  b=f2;
  c=f3;

  if (a>b){
    x=a;
    a=b;
    b=x; 
  }

  if (b>c){
    b=c;
  }
  if (a>b){
    return a;
  }
  return b;
}
//------------------------------------------------------------------------------------------------------------ 
U16 deltaT_ds(U16 base){ //Verstrichene Zeit in Dezisekunden (1/10s) bestimmen (berlauf nach 50 Minuten)
//------------------------------------------------------------------------------------------------------------ 
//Aufruf mit 0xffff liefert Referenzwert (Start der Zeitmessung); 
//Aufruf mit Referenzwert liefert seit Start vergangene Zeit in 1/10 s
  if (0xffff==base) return g.ms100;
  if (g.ms100 >= base){
    return (g.ms100-base);
  }else{
    return (g.ms100 + 30000L - base);
  }
}


#if 0

//------------------------------------------------------------------------------------------------------------ 
ISR(TIM1_OVF_vect){  //Alle 64s (PWM-Timer)
//------------------------------------------------------------------------------------------------------------ 
static U8 x;  
  x++;
  if (!x){
    PORTB |= BIT2;  //alle 16ms
  }
}
#else
//Die gleiche Funktionalitt handoptimiert in Assembler
//------------------------------------------------------------------------------------------------------------ 
void __attribute__ ((naked)) TIM1_OVF_vect(void){ //Alle 64s (PWM-Timer)
//------------------------------------------------------------------------------------------------------------ 
  asm("in	r16, 0x3f");  // Status merken (R16 am Dateianfang reserviert)
  asm("inc r17"); // Zhler inkrementieren
  asm("brne	.+2"); //Nicht NUll? Portnderung berspringen.
  asm("sbi	0x18, 2");   // High auf PB2 ausgeben - passiert nach 16ms
  asm("out	0x3f, r16"); // Status wiederherstellen
  asm("reti");
}
#endif


#if 0
//Registeradressen aus *.lss Datei fischen um diesen Code in Assembler zu schreiben
//------------------------------------------------------------------------------------------------------------ 
ISR(TIM0_COMPA_vect){ // 64KHz (ACO Timer)
//------------------------------------------------------------------------------------------------------------ 
  if (ACSR & (1<<ACO)){
    PORTB |= BIT0;
  }else{
    PORTB &=~BIT0;
  }  
}
#else
//Die gleiche Funktionalitt handoptimiert in Assembler
//------------------------------------------------------------------------------------------------------------ 
void __attribute__ ((naked)) TIM0_COMPA_vect(void){ // 64KHz (ACO Timer)
//------------------------------------------------------------------------------------------------------------ 
  asm("in	r16, 0x3f"); // Status merken (R16 am Dateianfang reserviert)
  asm("sbic	0x08, 5");   // Komparator abfragen (ACO)
  asm("rjmp	.+6");
  asm("cbi	0x18, 0");   // LOW auf PB0 ausgeben
  asm("out	0x3f, r16"); // Status weiderherstellen
  asm("reti");

  asm("sbi	0x18, 0");   // High auf PB0 ausgeben
  asm("out	0x3f, r16"); // Status wiederherstellen
  asm("reti");
}
#endif

//------------------------------------------------------------------------------------------------------------ 
U8 mppRegler1(U16 Umodul){ 
//------------------------------------------------------------------------------------------------------------ 
/*
   - Schaltet Schaltwandler aus um Modul im Leerlauf zu betreiben.
       (das klappt nicht ganz, da der Prozessor weiterhin Strom zieht)
   - Wartet, bis Modulspannung nicht mehr steigt (und Mindestwert berschreitet)
   - Mit Leerlaufspannung und nimmt 4/5 davon als MPP-Spannung


   Liefert 0 solange der Vorgang andauert, 1 wenn MPP eingestellt.

*/
static U8  step=0;
static U16 Umerk=0;
U16 warteDs=0;

  switch(step)
  {
    case 0: //Leerlaufspannung messen  
      TIMSK0=0; //Interrupts fr Komparatorauswertung aus
      OCR1A = 511; //Hchste mgliche Spannung anfordern 
      PORTB &=~BIT0; //FET aus

      Umerk = Umodul;
      warteDs=deltaT_ds(0xffff);
      if (Umodul > 5000L){  // (Mindestwert: knapp 1/3 Leerlaufspannung. Bei 24V Leerlauf: 7V)
        step++;
      }
      break;

    case 1: //warte bis Spannung nicht mehr steigt. Prfe alle 800ms
      if (deltaT_ds(warteDs) >= 8){
        warteDs=deltaT_ds(0xffff);
        if (Umerk >= Umodul) step++; //Spannung nicht gestiegen
        Umerk = Umodul;
      }
      break;

    case 2: //Rate MPP; Starte PWM
      // Umpp  = Umodul *4 /5     Umodul:0..1023*16;  PWM:0..511 (also 1/32stel von Umodul)
      // OCR1A = Umpp/32;  *4/5/32 => *4/160 => /40
      TIMSK0=(1<<OCIE0A); //Interrupts wieder ein
      OCR1A = Umodul/40; 
      warteDs=deltaT_ds(0xffff);
      step++;      
      break; //Habe nun geratenen MPP

    case 3: //Warte 2s bis MPP eingestellt
      if (deltaT_ds(warteDs) >= 20){
        step=0;         
        return 1;
      }
      break;

    default: 
      step=0;
      break;
  }
  return 0;
}
//------------------------------------------------------------------------------------------------------------ 
U8 mppRegler2(U16 Pmodul, U16 Umodul){ 
//------------------------------------------------------------------------------------------------------------ 
// merkt sich Pmodul, Aendert USoll, und bewertet beim nchsten Aufruf neues Pmodul

  //Ausgehend von momentanem USoll:
  // Laufe mit USoll max. MAX_SEARCH_STEPS nach oben, bis Leistung schlechter wird.
  // Springe auf Anfangswert
  // dann max MAX_SEARCH_STEPS nach unten bis Leistung schlechter wird.
  // Neues USoll ist die Mitte der beiden gefundenen Grenzen.
  // Ob Leistung schlechter geworden ist, wird anhand der letzten 2 Messwerte bestimmt.
  // Nur wenn beide schlechter geworden sind, wird die Suche eingestellt.

//Liefert 
// -1 wenn ein Regelzyklus noch nicht beendet
//  1 wenn beim Regeln an eine Grenze gestossen (Regelung fortfuehren empfohlen)
//  2 wenn an beide Grenzen gestossen (Bin irgendwo. ggf. mal 4/5*Leerlauf nehmen)
//  0 wenn beide Grenzen nicht erreicht (Scharf umgrenzten MPP prima gefunden)
/*
  0 wenn ein Regelzyklus noch nicht beendet
  1 wenn beide Grenzen nicht erreicht und Usoll um max. 1 gendert 
         (Scharf umgrenzten MPP prima gefunden oder sinkende Einstrahlung)
  2 wenn beide Grenzen nicht erreicht und Usoll um mehr als 1 gendert
         (MPP leicht korrigiert oder beginnende sinkende Einstrahlung)
  3 wenn beim Regeln an eine Grenze gestossen (Regelung fortfuehren empfohlen)
  4 wenn an beide Grenzen gestossen 
     Messung unbrauchbar da steigende Einstrahlung oder Leistungskurve noch zu flach
    (wenn lnger andauert: ggf. mal MPP mit 4/5*Leerlauf bestimmen)

  Nach nderung der Sollspannung wird bis zu 320ms gewartet bis sich die Spannung
  eingestellt hat. Mittels TargetH und TargetL wird ein Bereich angegeben innerhalb dessen
  die Zielspannung erwartet wird. Da ein PWM Schritt rechnerisch 32 ADC Schritte entspricht,
  die beiden aber niemals exakt sind, wird der Zielbereich etwas grer (8 Zhler) gefat.
*/

#define MAX_SEARCH_STEPS 12
static U16 schrittweite=2;

static U16 Pstart=0;
static U16 Ustart=0;
static U8  stufe=0;

static U16 pwmStart=0;
static U16 pwmHigh=0;
static U16 pwmLow=0;

static U8  stepCount=0;
static U8  oberGrenzeErreicht=0;
static U8  unterGrenzeErreicht=0;
static U16 p1,p2; //Eimerkette der Leistung
static U8  ms32cnt=0;
static U16 Utarget=0; //Spannung die bei MPP-Regelung erreicht werden soll
static U8  rising=0; 
U16 pwm;

  ms32cnt++; 
  //Warte max. 10*32ms = 320ms bis Zielspannung erreicht 
  if (ms32cnt<8) {
    if (rising){
      if (Umodul >= (Utarget-10)) //PWM wird nie ganz genau wie ADC sein. Spendiere Toleranz.
        ms32cnt=9; //Ziel erreicht. Im nchsten Zyklus auswerten.
    }else{ //falling
      if (Umodul <= (Utarget+10)) //mit Toleranz.
        ms32cnt=8; //Ziel erreicht. Im bernchsten Zyklus auswerten.
    }
  }
  if (ms32cnt<10) return 0;

  ms32cnt=0; 

  pwm = OCR1A;

  switch(stufe)
  {
    case 0: //SuchStart
      Pstart = Pmodul-3; //-3 damit "schlechteres P" etwas deutlicher ist
      Ustart = Umodul;
      p1=p2 = Pstart;
      pwmStart = pwm;
      pwmLow   = pwm;
      pwmHigh  = pwm;
      oberGrenzeErreicht=0;
      unterGrenzeErreicht=0;
      stepCount=0;
      pwm += schrittweite;
      rising=1;
      Utarget=Ustart + (pwm - pwmStart)*32;
      stufe=1; // Suche nach oben
      break; 

    case 1: // Suche nach oben
      p2=p1;
      p1=Pmodul;
      if (  //Bleibt P auch nach zwei Messungen schlechter?
                (p1 <  Pstart)
             && (p2 <  Pstart)
         )
      {
        stufe=2;
      }else{
        pwmHigh = pwm; //Merke erreichte Obergrenze
      }
      pwm += schrittweite;
      if (pwm>511) pwm=511;
      stepCount++;
      if (stepCount > MAX_SEARCH_STEPS){
        oberGrenzeErreicht=1;
        stufe=2;
      }
      rising=1;
      Utarget=Ustart + (pwm- pwmStart)*32;
      break;

    case 2:  //Springe zurueck auf Anfangswert und einen Schritt nach unten.
      pwm = pwmStart - schrittweite;
      stepCount=1;
      rising=0;
      Utarget=Ustart - (pwmStart-pwm)*32;
      stufe=3;
      break;

    case 3: // Warte, bis Hardware USoll-Sprung eingestellt hat
      p1=p2= Pstart;
      stufe = 4;
      break;

    case 4: // Suche nach unten
      p2=p1;
      p1=Pmodul;
      if (//Bleibt P auch nach zwei Messungen schlechter?
             ((p1+2) <  Pstart)
          && ((p2+2) <  Pstart)
         )
      {
        stufe=5;
      }else{
        pwmLow = pwm; //Merke erreichte Untergrenze
        
        if (pwm > 130) pwm -= schrittweite; //Begrenze nach unten , um dem Prozessor nicht den Saft abzudrehen.
                                         //(130 entspricht gut 6V bei Spannungsteiler ausgelegt auf 24V Modulleerlaufspannung)
        stepCount++;
        if (stepCount > MAX_SEARCH_STEPS){
          unterGrenzeErreicht=1;
          stufe=5;
          break;
        }
      }
      rising=0;
      Utarget=Ustart - (pwmStart-pwm)*32;
      break;

    case 5: //Auswertung
      //Nimm die Mitte aus pwmHigh und pwmLow
      pwm = (pwmLow + pwmHigh) /2;
      if (pwm > pwmStart){
        rising=1;
      }else{
        rising=0;
      }
      Utarget=Ustart + pwm*32 - pwmStart*32;
      stufe=6;
      break;

    case 6: //Warten bis neuer MPP eingestellt
      stufe=0; //Bin fertig. Beginne neuen Zyklus.
      break;

    default:
      stufe=0;
      break;
  }
  OCR1A = pwm;

  if (0==stufe){ //Zyklus beendet
   if (unterGrenzeErreicht && oberGrenzeErreicht){ //Messung unbrauchbar
     schrittweite =2;
     return   4;  //Beide Grenzen
   }
   if (!unterGrenzeErreicht && !oberGrenzeErreicht){//MPP gefunden
     schrittweite =1;
     if ((pwmStart>pwm+1) || (pwmStart < pwm-1)) return 2; //Keine Grenze
     return 1; //Keine Grenze, kaum nderung
   }
   //MPP weit verschoben
   schrittweite =3;
   return 3;   //Eine Grenze
  }
  return 0; //bin am Regeln
} // mppRegler2

//------------------------------------------------------------------------------------------------------------ 
void sendByte(U8 data){ //Byte auf serieller Schnittstelle ausgeben
//------------------------------------------------------------------------------------------------------------ 
  //Startbit(Low), LSB...MSB, Stopbit(High)  
  //Format: 9600 Baud, 8N1

//handgemachter USART - USI kostet nur zustzlichen Strom und bringt keine Erleichterung da ich 
//keinen Timer ber habe und der Takt somit auch per SW gemacht werden muss.

#define TX_HIGH PORTB|= BIT1
#define TX_LOW  PORTB&=~BIT1
#define DELAY_1_BIT _delay_us(86) //Eigentlich 104. Da aber viele Interrupts dazwischenkommen deutlich weniger.
#define DELAY_15_BIT _delay_us(150)
U8 n;
  n=8;  
  TX_LOW;  //Startbit
  while(n--){
    DELAY_1_BIT;
    if (data & 1) TX_HIGH; else TX_LOW;
    data>>=1;
  }
  DELAY_1_BIT;
  TX_HIGH; //    1.5 Stoppbits
  DELAY_15_BIT; 
}

//------------------------------------------------------------------------------------------------------------ 
void sendNumber3(U16 n){ //Zahl 3stellig auf serieller Schnittstelle ausgeben
//------------------------------------------------------------------------------------------------------------ 
  sendByte(n/100 + '0');
  n%=100;
  sendByte(n/10 + '0');
  n%=10;
  sendByte(n+ '0');
}

//------------------------------------------------------------------------------------------------------------ 
void sendNumber5(U16 n){ //Zahl 5stellig auf serieller Schnittstelle ausgeben
//------------------------------------------------------------------------------------------------------------ 
  sendByte(n/10000 + '0');
  n%=10000;
  sendByte(n/1000 + '0');
  n%=1000;
  sendNumber3(n);
}

//------------------------------------------------------------------------------------------------------------ 
int main(void){
//------------------------------------------------------------------------------------------------------------ 

static U8  step=0;
static U8  suchergebnis;
static U16 wartezeit=10;
static U8  nichtGefunden=0;
static U16 warteDs=0;
static U16 Umodul, Umerk;
static U16 Imodul;
static U16 Pmodul;
static U16 PmodulRef=0;
static U16 tic10000=0;
static U16 p1,p2,p3,p4;
U32 tmp;
U16 y;
U8  i;
U8  cycleTime=0;
U8  phase=0;
U8  txStep=0;
U8  changeCount=0;

  cli();
  DDRA = BIT6;      //PWM fuer MPP-Spannung 
  PORTA= BIT4|BIT5|BIT6|BIT7; //Sollspannung so hoch wie mglich.Pull-Ups auf unbenutzte Eingnge

  DDRB = BIT0|BIT1; //FET Tiefsetzsteller, Tx serielle Schnittstelle
  PORTB= BIT2|BIT3; //FET aus, Pull-Ups auf unbenutzte Eingnge



  ADMUX =(1<<REFS1) ; //Uref=1.1, input on ADC0
  ADCSRA=(1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) ; // 8MHz/64 gibt 125KHz
  ADCSRB=0;
  DIDR0 = (1<<ADC0D)|(1<<ADC1D)|(1<<ADC2D)|(1<<ADC3D); //Digitale Eingnge abschalten

  //Timer 0: Interrupts fr Komparator (64KHz)
  TCCR0A=(1<<WGM01);  //CTC
  TCCR0B=(1<<CS00);   //Teiler1(8MHz)
  OCR0A = 124; // Teiler / 125 - > 64KHz
  OCR0B = 0; 
  TIMSK0=(1<<OCIE0A);
  
  //Timer1: 9Bit PWM (15625Hz)
  TCCR1A=(1<<WGM11)|(1<<COM1A1); //WGM Mode 6: Fast PWM, non-inverting mode on A
  TCCR1B=(1<<WGM12)|(1<<CS10);   //WGM Mode 6, Teiler=1 (8MHz)
  OCR1A = 511; //Erstmal den FET mglichst aus
  OCR1B = 0; 
  TIMSK1= (1<<TOIE1);

  ACSR  =0; //Analogkomparator ein
  PRR   =(1<<PRUSI); // USI aus (Strom sparen)

  memset((U8 *)&g, sizeof(Tglob),0);
  Umerk=0;

  _delay_ms(2000); //2s warten bis Elkos geladen
  
  sei();
  
  while(1){

    //Zykluszeit 32ms  - warte bis die rum sind.

    //Timer setzt alle 16ms PB2 (genau:16,384ms)
    if (!(PORTB & BIT2)) continue;
    PORTB &=~ BIT2;
    if (phase){
      phase=0; 
      continue;
    }
    phase=1;
 
    //Hier komme ich alle 32ms hin. (genau: 32.768)
    g.ms32++;

    tic10000+=3277L;  //3276.8
    if (tic10000>=10000L){
      tic10000-=10000L;
      g.ms100++;
      if (g.ms100 >= 30000L) g.ms100 = 0;
    }

    //Umodul und Imodul 16x messen. Dauert ca. 4ms
    //Die ersten beiden Messungen dienen zum fllen des Median-Filters.
    ADMUX=(1<<REFS1)|0b001110 ; //Uref=1.1, (ADC1 - ADC3)x1 <--Umodul
    Umodul = 0;
    Imodul = 0;
    for (i=0; i<18; i++){ //2+16 Messungen
      ADCSRA |= (1<<ADSC); //Start conversion
      while (ADCSRA & (1<<ADSC))
      ;  //wait until ADC has finished
      y = ADC;
      if (i > 1){
        Umodul += medianU(y);
      }else{ //Nur den Filter fllen
        medianU(y);
      } 
    }     
    ADMUX=(1<<REFS1)|0b001011 ; //Uref=1.1, (ADC0 - ADC3)x20 <--Imodul
    for (i=0; i<18; i++){ //2+16 Messungen
      ADCSRA |= (1<<ADSC); //Start conversion
      while (ADCSRA & (1<<ADSC))
      ;  //wait until ADC has finished
      y = ADC;
      if (i > 1){
        Imodul += medianI(y);
      }else{ //Nur den Filter fllen
        medianU(y);
      } 
    }
    //Schon mal den MUX zurckschalten um den S+H Kondensator zu laden
    //(Umodul ist ja recht hochohmig)
    ADMUX=(1<<REFS1)|0b001110 ; //Uref=1.1, (ADC1 - ADC3)x1 <--Umodul


    //--------------- Jetzt habe ich aktuelle Messwerte -----------------

    //Wenn Eingangsspannung zu niedrig (unter 7V bei 24V  Modulleerlaufspannung): 
    //FET abschalten, MPP-Suche verhindern.
    if (Umodul < 4774L){  // 16*1023 * (7/24)
      step=0;
    }

    //Fifo der Leistung
    p4=p3;
    p3=p2;
    p2=p1;
    p1=Pmodul;

    //Momentanleistung P=U*I. Etwas jongliererei um Interger-Overflow zu verhindern.
    tmp = Umodul; //Typwandlung U16->U32
    tmp*= Imodul; //Jetzt kann man ohne berlauf multiplizieren
    tmp/= 16384L; //die unteren Bits sind rauschen. Weg damit! 
                  //Jetzt ist der Wertebereich 0..16352  (16352=16*1023*16*1023/16384)
    Pmodul = tmp; //Das passt prima in einen U16 
 
    switch(step){
      case 0: //MPP mittels Leerlaufspannung raten
        if (mppRegler1(Umodul)){
          //Habe nun geratenen MPP
          step++;  
        }
        break;

      case 1:  //Mpp suchen
        suchergebnis = mppRegler2(Pmodul, Umodul);
        switch (suchergebnis){
          case 0:  break; //Bin noch am suchen
          case 1: // MPP gefunden. in 30s wieder suchen
            step++; 
            wartezeit = 300; 
            nichtGefunden=0; 
            break; 
          case 2: // MPP leicht verschoben. In 10s wieder suchen
            step++; 
            wartezeit = 100; 
            nichtGefunden=0; 
            break; 
          case 3: // MPP verschoben. Gleich weitersuchen.
            step++; 
            wartezeit =   0; 
            nichtGefunden=0; 
            break; 
          case 4: // Unbrauchbar. Warte kurz (Einstrahlung ndert sich grade)
            step++; 
            wartezeit =  60; 
            nichtGefunden++; 
            if (nichtGefunden > 20){ //MPP Suche aussichtslos. Beginne nochmal ganz von vorn.
              step = 0;
            }
            break; 
        }
        if (suchergebnis){
          warteDs = deltaT_ds(0xffff);
        }
        break;


      case 2: //Warte bis neuer MPP eingestellt
        if (deltaT_ds(warteDs) > 20){  //2 Sekunden sind rum
          //Merke Momentanleistung
          PmodulRef = Pmodul;
          changeCount=0;
          warteDs = deltaT_ds(0xffff);
          step++;
        }
        break;

      case 3: //Warte bis nchste MPP-Suche dran ist oder Einstrahlung sich ndert
        if (deltaT_ds(warteDs)+30 > wartezeit){ //30 sind schon in Step5 vergangen
          step++;
        }else{
          //MPP-Suche anstoen wenn mehrfach ber 10% Leistungnderung entdeckt 
          if (
                  (Pmodul > (PmodulRef/10*11)) 
               || (Pmodul < (PmodulRef/10* 9))
             )
          {  //Leistung verndert
            changeCount++;
            if (changeCount > 10) step++;  //Neue Suche starten
          }else{
             if (changeCount) changeCount--;
          } 
        }
        break;
      case 4: //Warte wenn sich die Einstrahlung grade stark ndert (ca 0.4% in 150ms)
        if (Pmodul>=p4) {
          y = Pmodul - p4;
        }else{
          y = p4 - Pmodul;
        }
        if ((y < (Pmodul / 256)) || (y<16)) step=1;
        break;

      default:
        step=0;
        break;
    }//switch

#if 1
    //Messwerte ca. alle 400ms (200ms + 6*32ms Sendezeit) auf serieller Schnittstelle ausgeben
    //Da ein Zyklus nur 16ms dauert mssen die Daten portionsweise ausgegeben werden.
    if (deltaT_ds(cycleTime) >=2){
      switch(txStep++){
       case 0:
         sendByte(13);
         sendByte(10);
         sendByte('P');
         sendByte('=');
         break;
       case 1:
        sendNumber5(Pmodul);
         break;
       case 2:
         sendByte(' ');
         sendByte('I');
         sendByte('=');
         break;
       case 3:
        sendNumber5(Imodul);
         break;
       case 4:
         sendByte(' ');
         sendByte('S'); //Sollwert (PWM)
         sendByte('=');
         break;
       case 5:
         sendNumber3(OCR1A);
         cycleTime=deltaT_ds(0xffff);
         txStep=0;
       break;
      }
    }
#endif

  }//while(1)
}
//EOF
